<?php declare(strict_types=1);
/**
 * This file is part of assert-function-signature.
 */

namespace Grifart\AssertFunction;

final class SignatureAssertionUtil
{
	private function __construct()
	{
	}

	public static function checkSignature(callable $function, array $parameters, ?string $expectedReturnType): void
	{
		$reflection = new \ReflectionFunction($function);

		// NUMBER OF PARAMETERS CHECK:
		self::checkParameters($parameters, $reflection);

		// RETURN TYPE:
		self::checkReturnType($expectedReturnType, $reflection);
	}

	// parameters:
	private static function checkParameters(array $parameters, \ReflectionFunction $reflection): void
	{
		$numberOfParameters = $reflection->getNumberOfParameters();
		if ($numberOfParameters !== count($parameters)) {
			throw FunctionSignatureAssertionError::wrongNumberOrArguments($reflection, count($parameters), $numberOfParameters);
		}

		// PARAMETER TYPES CHECK:
		$i = 0;
		/** @var string[] $parameters */
		foreach ($parameters AS $parameter) {
			$parameterReflection = $reflection->getParameters()[$i++];
			assert($parameterReflection instanceof \ReflectionParameter);

			if ($parameter === 'any') continue;
			assert(is_string($parameter));

			self::checkFunctionParameter(
				$reflection,
				$parameterReflection,
				...self::parseType($parameter)
			);
		}
	}

	private static function checkFunctionParameter(
		\ReflectionFunction $functionReflection,
		\ReflectionParameter $parameterReflection,
		string $expectedParameterType,
		bool $expectedNullable
	): void
	{
		$reflectionType = $parameterReflection->getType();
		if ($expectedParameterType === 'mixed') {
			// must accept everything
			if ($reflectionType === NULL) {
				return;
			}

			throw FunctionSignatureAssertionError::parameterDoesNotAcceptAllRequiredValues(
				$functionReflection,
				$reflectionType
			);
		}
		assert($reflectionType instanceof \ReflectionType);

		// checks:
		$actuallyNullable = $reflectionType->allowsNull();
		if ($expectedNullable !== $actuallyNullable) {
			throw FunctionSignatureAssertionError::parameterNullability(
				$functionReflection,
				$parameterReflection,
				$expectedNullable,
				$actuallyNullable
			);
		}
		if ($expectedParameterType !== (string) $reflectionType) {
			throw FunctionSignatureAssertionError::wrongParameterType(
				$functionReflection,
				$parameterReflection,
				$expectedParameterType,
				$reflectionType
			);
		}
	}

	// Return-type:
	private static function checkReturnType(?string $expectedReturnType, \ReflectionFunction $reflection): void
	{
		if ($expectedReturnType === NULL || $expectedReturnType === 'mixed') {
			return;
		}

		if (!$reflection->hasReturnType()) {
			throw FunctionSignatureAssertionError::missingReturnType($reflection, $expectedReturnType);
		}

		[$expectedReturnType, $expectingNullable] = self::parseType($expectedReturnType);
		$returnTypeReflection = $reflection->getReturnType();

		$actuallyNullable = $returnTypeReflection->allowsNull();
		if ($expectingNullable !== $actuallyNullable) {
			throw FunctionSignatureAssertionError::wrongReturnTypeNullability(
				$reflection,
				$expectingNullable,
				$actuallyNullable
			);
		}
		if ($expectedReturnType !== (string) $reflection->getReturnType()) {
			throw FunctionSignatureAssertionError::wrongReturnType(
				$reflection,
				$expectedReturnType,
				(string) $reflection->getReturnType()
			);
		}
	}

	// helpers:

	private static function parseType(string $type): array
	{
		$cleanType = $type; // e.g. ?Namespace\Class
		$nullable = FALSE;
		if ($type[0] === '?') {
			$nullable = TRUE;
			$cleanType = substr($type, 1); // without leading '?'
		}
		return [$cleanType, $nullable];
	}

}
